// fldDynaNR.cc

//----------------------------------------------------------------------
//	FldDynaNR		$B2OF;LV%b%G%k(B
//----------------------------------------------------------------------

#include "fldDynaNR.h"	// $B%/%i%9(B FldDynaNR $B$NDj5A(B
#include "grid.cc"

Boolean FldDynaNR::Init(char* o_name, int o_num, size_t num_row, size_t
					  num_col, char* f_grid, char* f_all, char* f_cell,
						SwitchFlag sf)
{
	int i, j, k, l, m;
	size_t r, c;
	char str[1024];

	if ((int)num_row < 0 || (int)num_col < 0) {
		Eprint(__FILE__, __LINE__);
		// Invalid line. (num_row < 0 or num_col < 0)
		// FldDynaNR::Init
		Oprint(); exit(1);
	}
	if (sf != RUNOFF && sf != VARIED) {
		Eprint(__FILE__, __LINE__);
		// Invalid flag. SwitchFlag is RUNOFF or VARIED.
		// FldDynaNR::Init
		Oprint(); exit(1);
	}

	n_row = num_row;
	n_col = num_col;
	n_cell = n_row * n_col;
	sflag = sf;

	bc.Resize(n_row, n_col);
	position.Resize(n_row, n_col);
	if (sflag == VARIED) {
		bp_in.Resize(n_row, n_col);
		bp_out.Resize(n_row, n_col);
		sn_up.Resize(n_row, n_col);
		sn_low.Resize(n_row, n_col);
	}

	FILE* fp;

	if ((fp = fopen(f_grid, "r")) == NULL) {
		Eprint(__FILE__, __LINE__);
		// Cannot open file.
		// FldDynaNR::Init
		Oprint(); exit(1);
	}

	if (fgets(str, 1024, fp) != NULL) {
		if (sscanf(str, "%d", &n_parts) != 1) {
			Eprint(__FILE__, __LINE__);
			// Invalid inputs.
			// FldDynaNR::Init
			Oprint(); exit(1);
		}
		if (n_parts < 0) {
			Eprint(__FILE__, __LINE__);
			// Invalid inputs. (n_parts < 0)
			// FldDynaNR::Init
			Oprint(); exit(1);
		}
	}
	else {
		Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
	}
	
	cn = new CellNumber[n_parts];

	for (i = 0; i < n_parts; i++) {
		if (fgets(str, 1024, fp) != NULL) {
			if (sscanf(str, "%d %d", &j, &k) != 2) {
				Eprint(__FILE__, __LINE__);
				Oprint(); exit(1);
			}
			cn[i].row = j - 1;
			cn[i].col = k - 1;
			if (j > (int)n_row || k > (int)n_col) {
				Eprint(__FILE__, __LINE__);
				Oprint(); exit(1);
			}
		}
		else {
			Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
		}
	}

	fclose(fp);

	if ((fp = fopen(f_all, "r")) == NULL) {
		Eprint(__FILE__, __LINE__);
		// Cannot open file.
		// FldDynaNR::Init
		Oprint(); exit(1);
	}

	n_is = 0;
	while (fgets(str, 1024, fp) != NULL)	// $B%U%!%$%k$r6uFI$_$9$k(B
		n_is++;
	
	fseek(fp, 0L, 0);	// $B%U%!%$%k%]%$%s%?$r@hF,$KLa$9(B

	bp = new BoundaryPoint[n_is];
	for (i = 0; i < n_is; i++) {
		bp[i].row.Resize(2);
		bp[i].col.Resize(2);
		bp[i].segnum.Resize(2);
	}
	int kk, ll, mm, nn, n;
	for (i = 0; i < n_is; i++) {	// $B2OF;LVA4BN$N@\B34X78$rFI$_<h$k(B
		if (fgets(str, 1024, fp) != NULL) {
			if (sscanf(str, "%d %d %d %d %d %d %d %d", &k, &l, &m,
					   &n, &kk, &ll, &mm, &nn) != 8) {
				Eprint(__FILE__, __LINE__);
				// Input file about connecting segments 
				// is invalid. (Cell number and segment number)
				// FldDynaNR::Init
				Oprint(); exit(1);
			}
			bp[i].row[0] = k - 1;
			bp[i].col[0] = l - 1;
			bp[i].segnum[0] = -n;
			bp[i].row[1] = kk - 1;
			bp[i].col[1] = ll - 1;
			bp[i].segnum[1] = nn;
		}
		else {
			Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
		}
	}

	if ((fp = fopen(f_cell, "r")) == NULL) {
		Eprint(__FILE__, __LINE__);
		// Cannot open file.
		// FldDynaNR::Init
		Oprint(); exit(1);
	}

	for (i = 0; i < (int)n_cell; i++) {	// $B3F%;%kFb$N>pJs$rFI$_<h$k(B
		if (fgets(str, 1024, fp) != NULL) {
			if (sscanf(str, "%d %d", &r, &c) != 2) {
				Eprint(__FILE__, __LINE__);
				// Input file about the information of each cell
				// is invalid. (The cell number)
				// FldDynaNR::Init
				Oprint(); exit(1);
			}
			bc(r - 1, c - 1).Connect_segment(fp);
		}
		else {
			Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
		}
	}

	// $BITEyN.$N7W;;$N$?$a$N=`Hw(B

	if (sflag == VARIED) {
		int up, low;
		for (r = 0; r < n_row; r++) {
			for (c = 0; c < n_col; c++) {
				bc(r, c).Re_num_of_bp(sn_up(r, c), sn_low(r, c));
				up = sn_up(r, c).Size();
				low = sn_low(r, c).Size();
				bp_in(r, c).row.Resize(up);
				bp_in(r, c).col.Resize(up);
				bp_in(r, c).segnum.Resize(up);
				bp_out(r, c).row.Resize(low);
				bp_out(r, c).col.Resize(low);
				bp_out(r, c).segnum.Resize(low);
				for (j = 0; j < up; j++) {
					for (k = 0; k < n_is; k++) {
						if ((size_t)bp[k].row[0] == r &&
							(size_t)bp[k].col[0] == c &&
							bp[k].segnum[0] == - sn_up(r, c)[j]) {
							bp_in(r, c).row[j] = bp[k].row[1];
							bp_in(r, c).col[j] = bp[k].col[1];
							bp_in(r, c).segnum[j] =
								bp[k].segnum[1]; break;
						}
						else if ((size_t)bp[k].row[1] == r &&
								 (size_t)bp[k].col[1] == c
								 && bp[k].segnum[1] == - sn_up(r, c)[j]) {
							bp_in(r, c).row[j] = bp[k].row[0];
							bp_in(r, c).col[j] = bp[k].col[0];
							bp_in(r, c).segnum[j] =
								bp[k].segnum[0]; break;
						}
					}
				}
				for (j = 0; j < low; j++) {
					for (k = 0; k < n_is; k++) {
						if ((size_t)bp[k].row[0] == r &&
							(size_t)bp[k].col[0] == c &&
							bp[k].segnum[0] == sn_low(r, c)[j]) {
							bp_out(r, c).row[j] = bp[k].row[1];
							bp_out(r, c).col[j] = bp[k].col[1];
							bp_out(r, c).segnum[j] =
								- bp[k].segnum[1]; break;
						}
						else if ((size_t)bp[k].row[1] == r &&
								 (size_t)bp[k].col[1] == c &&
								 bp[k].segnum[1] == sn_low(r, c)[j]) {
							bp_out(r, c).row[j] = bp[k].row[0];
							bp_out(r, c).col[j] = bp[k].col[0];
							bp_out(r, c).segnum[j] =
								- bp[k].segnum[0]; break;
						}
					}
				}
			}
		}
	}

	n_upper = n_lower1 = n_lower2 = 0;

	for (r = 0; r < n_row; r++) {
		for (c = 0; c < n_col; c++) {
			n_upper = n_upper + bc(r, c).Re_ur();
			n_lower1 = n_lower1 + bc(r, c).Re_lr1();
			n_lower2 = n_lower2 + bc(r, c).Re_lr2();
		}
	}

	rp_up = new Receive<double>[n_upper];
	rp_low = new Receive<double>[n_lower1];
	for (i = 0; i < n_upper; i++) {
		rp_up[i].Init("rp_up", i);
	}
	for (i = 0; i < n_lower1; i++) {
		rp_low[i].Init("rp_low", i);
	}

	sp = new Send<double>[(n_lower1 + n_lower2)];
	for (i = 0; i < (n_lower1 + n_lower2); i++) {
		sp[i].Init("sp", i);
	}

	int sum = 0;
	for (i = 0; i < n_parts; i++) {
		r = cn[i].row;
		c = cn[i].col;
		sum = sum + bc(r, c).Re_num_of_seg();
	}

	sp_q = new Send<Vdary>[sum];
	sp_h = new Send<Vdary>[sum];
	for (i = 0; i < sum; i++) {
		sp_q[i].Init("sp_q", i);
		sp_h[i].Init("sp_h", i);
	}

	return Element::Init(o_name, o_num, 1, NULL);	// $BD>@\DL?.$9$kAj<j(B
													// $B$O0l$D(B($BB&J}N.F~NL(B
													// $BMQ(B)
}

FldDynaNR::FldDynaNR(char* o_name, int o_num, size_t num_row, size_t
				 num_col, char* f_grid, char* f_all, char* f_cell,
					 SwitchFlag sf)
	: Element()
{
	class_name = "FldDynaNR";
	Init(o_name, o_num, num_row, num_col, f_grid, f_all, f_cell, sf);
}

FldDynaNR::FldDynaNR(void)
	: Element()
{
	class_name = "FldDynaNR";
}

void FldDynaNR::Register_receive_ports(void)
{
	int i;

	for (i = 0; i < n_upper; i++) {
		Register((void*)&rp_up[i]);
	}
	for (i = 0; i < n_lower1; i++) {
		Register((void*)&rp_low[i]);
	}
}

void FldDynaNR::Register_send_ports(void)
{
	int i;
	size_t r, c;

	for (i = 0; i < (n_lower1 + n_lower2); i++) {
		Register((void*)&sp[i]);
	}

	int sum = 0;
	for (i = 0; i < n_parts; i++) {
		r = cn[i].row;
		c = cn[i].col;
		sum = sum + bc(r, c).Re_num_of_seg();
	}

	for (i = 0; i < sum; i++) {
		Register((void*)&sp_q[i]);
		Register((void*)&sp_h[i]);
	}
}

void FldDynaNR::Set_parameter(FILE* fp)
{
	int i, j, k, l, num;
	int sum = 0;
	double v, w, z;
	size_t r, c;
	Viary p;
	char str[1024];

	// $B7W;;C<E@$N?t$rG[Ns$K5-21$9$k(B
	// $B7W;;C<E@$N?t$O(B 0, 5, 10, 15 $B!D(B $B$N$h$&$K2C;;$7$?$b$N$r(B
	// $B5-21$9$k(B($B$O$8$a$NCM$O(B 0)

	n_max = 0;

	for (r = 0; r < n_row; r++) {
		for (c = 0; c < n_col; c++) {
			position(r, c) = sum;
			i = bc(r, c).Re_comp_node();
			sum = sum + i;
			if (n_max < i) n_max = i;
		}
	}

	if (n_max <= 0) {
		Eprint(__FILE__, __LINE__);
		Oprint(); exit(1);
	}

	n_cn = sum;

	//	$BMWAG%b%G%k<+?H$N%Q%i%a%?$N@_Dj(B

	M_all = new double*[n_cn * 2];
	M_fixed = new double*[n_cn * 2];
	for (i = 0; i < n_cn * 2; i++) {
		M_all[i] = new double[n_cn * 2];
		M_fixed[i] = new double[n_cn * 2];
	}

	F_all.Resize(n_cn * 2);
	x_all.Resize(n_cn * 2);

	M_cell = new double*[n_max];
	for (i = 0; i < n_max; i++) {
		M_cell[i] = new double[n_max * 2];
	}

	F_cell.Resize(n_max);

	// $B7W;;;~4V4V3V!&N.NL5Z$S?e?<$NBeI=CM$rFI$_9~$`(B

	if (fgets(str, 1024, fp) != NULL) {
		if (sscanf(str, "%ld %lf %lf", &t_step, &Qlevel, &Hlevel) != 3)	
			{
				Eprint(__FILE__, __LINE__);
				// Cannot get time step or Qlevel or Hlevel.
				// FldDynaNR::Set_parameter
				Oprint(); exit(1);
			}
	}
	else {
		Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
	}

	predt = t_step;

	// Q-h $B4X78<0$K$*$1$k(B a $B$H(B b $B$rFI$_9~$`(B

	qhr = new QHRelation[n_lower2];
	for (i = 0; i < n_lower2; i++) {
		if (fgets(str, 1024, fp) != NULL) {
			if (sscanf(str, "%d %d %d %lf %lf %lf", &r, &c, &j, &v, &w,
					   &z) != 6) {
				Eprint(__FILE__, __LINE__);
				// Cannot get a and b in Q-H relationship.
				// FldDynaNR::Set_parameter
				Oprint(); exit(1);
			}
			else {
				qhr[i].row = r - 1;
				qhr[i].col = c - 1;
				qhr[i].segnum = j;
				qhr[i].a = v;
				qhr[i].b = w;
				qhr[i].zlow = z;
			}
		}
		else {
			Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
		}
	}

	//	$B3F%;%k$N%Q%i%a%?$r3F%;%k$KEO$9(B

	for (i = 0; i < (int)n_cell; i++) {
		if (fgets(str, 1024, fp) != NULL) {
			if (sscanf(str, "%d %d", &r, &c) != 2) {
				Eprint(__FILE__, __LINE__);
				fprintf(stderr, "cell number = %d %d\n", r, c);
				// Cannot get cell number in parameter file.
				// FldDynaNR::Set_parameter
				Oprint(); exit(1);
			}
			if (bc(r - 1, c - 1).Set_parameter(fp, t_step) != YES) {
				Eprint(__FILE__, __LINE__);
				// Cannot set parameter.
				// FldDynaNR::Set_parameter
				fprintf(stderr, "cellnumber = %d %d", r, c);
				Oprint(); exit(1);
			}
		}
		else {
			Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
		}
	}

	// $B8GDjItJ,$N78?t(B($B>eN.C<6-3&!&2<N.C<6-3&(B 1)$B$r9TNs$K5-21$9$k(B

	i = n_cn;

	for (j = 0; j < 2 * n_cn; j++) {		// $B9TNs$N=i4|2=(B
		for (k = 0; k < 2 * n_cn; k++) {
			M_fixed[j][k] = 0.;
		}
	}

	for (r = 0; r < n_row; r++) {			// $B6-3&>r7o$K4X$9$kItJ,(B
		for (c = 0; c < n_col; c++) {
			bc(r, c).Re_ur(p);
			k = position(r, c);
			num = p.Size();			// num : $B:G>eN.C<6-3&$N?t(B
			for (j = 0; j < num; j++) {
				l = bc(r, c).Re_position_of_variable(p[j], UPPER,
													 DISCHARGE);
				M_fixed[i][k * 2 + l] = 1.;
				i++;
			}

			bc(r, c).Re_lr1(p);
			num = p.Size();			// num : $B:G2<N.C<6-3&(B 1 $B$N?t(B
			for (j = 0; j < num; j++) {
				l = bc(r, c).Re_position_of_variable(p[j], LOWER,
													 DEPTH);
				M_fixed[i][k * 2 + l] = 1.;
				i++;
			}
		}
	}

	sum = 0;
	for (r = 0; r < n_row; r++) {
		for (c = 0; c < n_col; c++) {
			sum = sum + bc(r, c).Re_sum_of_cs();
		}
	}
	printf("The number of cross section is %d\n", sum);
}

void FldDynaNR::Set_initial_state(FILE* fp)
{
	int i;
	double lat;
	size_t r, c;
	char str[1024];


	// 	$B=i4|$N(B current_time $B$N@_Dj(B

	if (fgets(str, 1024, fp) != NULL) {
		if (sscanf(str, "%ld", &current_time) != 1) {
			Eprint(__FILE__, __LINE__);
			// Cannot set initial state.(current_time)
			// FldDynaNR::Set_initial_state
			Oprint(); exit(1);
		}
	}
	else {
		Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
	}

	// $B3F%;%k$N=i4|>uBVNL$N@_Dj(B

	for (i = 0; i < (int)n_cell; i++) {
		if (fgets(str, 1024, fp) != NULL) {
			if (sscanf(str, "%d %d", &r, &c) != 2) {
				Eprint(__FILE__, __LINE__);
				// Cannot set initial state.(cell number)
				// FldDynaNR::Set_initial_state
				Oprint(); exit(1);
			}
			if (((FldSlope*)object_array[0])->Re_lat(r - 1, c - 1,
													 current_time,
													 lat) != YES) {
				Eprint(__FILE__, __LINE__);
				// Cannot get lateral inflow at initial time.
				// FldDynaNR::Set_initial_state
				fprintf(stderr, "cellnumber = %d %d", r, c);
				Oprint(); exit(1);
			}

			if (sflag == RUNOFF) {	// $BN.=P7W;;$N=i4|>uBVNL$r@_Dj(B
				if (bc(r - 1, c - 1).Set_init_for_runoff(fp, lat) !=
					YES) {
					Eprint(__FILE__, __LINE__);
					// Cannot set initial state for runoff model.
					// FldDynaNR::Set_initial_state
					fprintf(stderr, "cellnumber = %d %d", r, c);
					Oprint(); exit(1);
				}
			}
			else if (sflag == VARIED) {	// $BITEyN.$N7W;;$N=i4|>uBVNL$r@_(B
										// $BDj(B
				if (bc(r - 1, c - 1).Set_init_for_varied(fp) != YES) {
					Eprint(__FILE__, __LINE__);
					// Cannot set initial state for varied flow.
					// FldDynaNR::Set_initial_state
					fprintf(stderr, "cellnumber = %d %d", r, c);
					Oprint(); exit(1);
				}
			}
		}
		else {
			Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
		}
	}
}

void FldDynaNR::Save_terminal_state(FILE* fp)
{
	size_t r, c;

	if (fp != NULL) {
		fprintf(fp, "%ld\n", current_time);
	}
	else {
		Eprint(__FILE__, __LINE__);
		Oprint(); exit(1);
	}

	for (r = 0; r < n_row; r++) {
		for (c = 0; c < n_col; c++) {
			fprintf(fp, "%d %d\n", r, c);
			if (bc(r, c).Save_terminal_state(fp) != YES) {
				Eprint(__FILE__, __LINE__);
				// Cannot save terminal state.
				// FldDynaNR::Save_terminal_state
				Oprint(); exit(1);
			}
		}
	}

}

Boolean FldDynaNR::Initial_output(void)
{
	int i, j, k, l, num;
	size_t r, c;
	Vdary Qlr;
	Vdary* Qall;
	Vdary* Hall;

	j = 0;

	for (r = 0; r < n_row; r++) {
		for (c = 0; c < n_col; c++) {
			num = bc(r, c).Re_lr1() + bc(r, c).Re_lr2();
			Qlr.Resize(num);
			bc(r, c).Re_lowest_q(Qlr);
			for (i = 0; i < num; i++) {
				sp[j].Send_data(current_time, Qlr[i]);
				j++;
			}
		}
	}

	if (j != (n_lower1 + n_lower2)) {
		Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
	}
		
	l = 0;

	for (i = 0; i < n_parts; i++) {
		r = (size_t)cn[i].row;
		c = (size_t)cn[i].col;

		j = bc(r, c).Re_num_of_seg();
		Qall = new Vdary[j];
		Hall = new Vdary[j];
		
		if (bc(r, c).Output_variable(Qall, Hall, CURRENT) == YES) {
			for (k = 0; k < j; k++) {
				sp_q[l].Send_data(current_time, Qall[k]);
				sp_h[l].Send_data(current_time, Hall[k]);
				l++;
			}
		}
		else {
			Eprint(__FILE__, __LINE__);
			// Cannot output variable.
			// FldDynaNR::Initial_output
			Oprint(); exit(1);
		}
		delete [] Qall;
		delete [] Hall;
	}

	return YES;
}

void FldDynaNR::Make_equation(void)
{
	//	$B3F%;%k$KO"N)J}Dx<0$rN)$F$5$;!"9gN.ItJ,$NA]$-=P$7$r$5$;$k(B

	int i, j, k, l, m, n, ll, mm, nn, num, sum = 0;
	size_t r, c, rr, cc;
	double q, h;
	Vdary data(3);
	Viary p;

	for (i = 0; i < n_cn * 2; i++) {		// $B9TNs$N=i4|2=(B
		F_all[i] = 0.;
		for (j = 0; j < n_cn * 2; j++) {
			M_all[i][j] = M_fixed[i][j];
		}
	}

	for (r = 0; r < n_row; r++) {			// $B3F%;%k$K4X$9$kItJ,(B
		for (c = 0; c < n_col; c++) {
			if (bc(r, c).Make_equation(time_step) != YES) {
				Eprint(__FILE__, __LINE__);
				fprintf(stderr, "cellnumber = %d %d", r, c);
				Oprint(); exit(1);
			}
			if (bc(r, c).Sweep_equation(M_cell, F_cell) != YES) {
				Eprint(__FILE__, __LINE__);
				fprintf(stderr, "cellnumber = %d %d", r, c);
				Oprint(); exit(1);
			}

			num = bc(r, c).Re_comp_node();
			for (i = sum; i < (sum + num); i++) {
				for (j = sum * 2; j < (sum + num) * 2; j++) {
					M_all[i][j] = M_cell[i - sum][j - sum * 2];
				}
				F_all[i] = F_cell[i - sum];
			}
			sum = sum + num;
		}
	}

	i = n_cn + n_upper + n_lower1;

	for (r = 0; r < n_row; r++) {
		for (c = 0; c < n_col; c++) {
			bc(r, c).Re_lr2(p);
			k = position(r, c);
			num = p.Size();			// num : $B:G2<N.C<6-3&(B 2 $B$N?t(B
			for (j = 0; j < num; j++) {
				l = bc(r, c).Re_position_of_variable(p[j], LOWER,
													 DISCHARGE);
				M_all[i][k * 2 + l] = 1.;
				Re_a_and_b(r, c, p[j], data);
				h = bc(r, c).Re_variable(p[j], LOWER, DEPTH);
				M_all[i][k * 2 + l + 1] = -2. * data[0] *
					(h + data[2] + data[1]);
				q = bc(r, c).Re_variable(p[j], LOWER, DISCHARGE);
				F_all[i] = data[0] * (h + data[2] + data[1]) *
					(h + data[2] + data[1]) - q;
				i++;
			}
		}
	}

	for (j = 0; j < n_is; j++) {	// $BE,9g>r7o$K4X$9$kItJ,(B($BN.NL(B)
		for (k = 0; k < 2; k++) {
			l = bp[j].segnum[k];
			r = (size_t)bp[j].row[k];
			c = (size_t)bp[j].col[k];
			if (l > 0) {		// $B$=$N%N!<%I$KN.F~$9$k;~(B
				m = position(r, c);
				n = bc(r, c).Re_position_of_variable(l, LOWER,
													 DISCHARGE);
				M_all[i][2 * m + n] = 1.;
				F_all[i] = F_all[i] - bc(r, c).Re_variable(l, LOWER,
														   DISCHARGE);
			}
			else if (l < 0) {	// $B$=$N%N!<%I$+$iN.=P$9$k;~(B
				m = position(r, c);
				n = bc(r, c).Re_position_of_variable(-l, UPPER,
													 DISCHARGE);
				M_all[i][2 * m + n] = - 1.;
				F_all[i] = F_all[i] + bc(r, c).Re_variable(-l, UPPER,
														   DISCHARGE);
			}
		}
		i++;
	}

	for (j = 0; j < n_is; j++) {	// $BE,9g>r7o$K4X$9$kItJ,(B($B?e?<(B)
		ll = bp[j].segnum[0];
		rr = (size_t)bp[j].row[0];
		cc = (size_t)bp[j].col[0];
		if (ll > 0) {		// $B$=$N%N!<%I$KN.F~$9$k;~(B
			mm = position(rr, cc);
			nn = bc(rr, cc).Re_position_of_variable(ll, LOWER, DEPTH);
			M_all[i][2 * mm + nn] = 1.;
			F_all[i] = F_all[i] - bc(rr, cc).Re_variable(ll, LOWER,
														 DEPTH);
			l = bp[j].segnum[1];
			r = (size_t)bp[j].row[1];
			c = (size_t)bp[j].col[1];
			if (l > 0) {	// $B$=$N%N!<%I$KN.F~$9$k;~(B
				m = position(r, c);
				n = bc(r, c).Re_position_of_variable(l, LOWER, DEPTH);
				M_all[i][2 * m + n] = - 1.;
				F_all[i] = F_all[i] + bc(r, c).Re_variable(l, LOWER,
														   DEPTH);
				i++;
			}
			else if (l < 0) {	// $B$=$N%N!<%I$+$iN.=P$9$k;~(B
				m = position(r, c);
				n = bc(r, c).Re_position_of_variable(-l, UPPER,
													 DEPTH);
				M_all[i][2 * m + n] = - 1.;
				F_all[i] = F_all[i] + bc(r, c).Re_variable(-l, UPPER,
														   DEPTH);
				i++;
			}
		}
		else if (ll < 0) {		// $B$=$N%N!<%I$+$iN.=P$9$k;~(B
			mm = position(rr, cc);
			nn = bc(rr, cc).Re_position_of_variable(-ll, UPPER, DEPTH);
			M_all[i][2 * mm + nn] = 1.;
			F_all[i] = F_all[i] - bc(rr, cc).Re_variable(-ll, UPPER,
														 DEPTH);
			l = bp[j].segnum[1];
			r = (size_t)bp[j].row[1];
			c = (size_t)bp[j].col[1];
			if (l > 0) {	// $B$=$N%N!<%I$KN.F~$9$k;~(B
				m = position(r, c);
				n = bc(r, c).Re_position_of_variable(l, LOWER, DEPTH);
				M_all[i][2 * m + n] = - 1.;
				F_all[i] = F_all[i] + bc(r, c).Re_variable(l, LOWER,
														   DEPTH);
				i++;
			}
			else if (l < 0) {	// $B$=$N%N!<%I$+$iN.=P$9$k;~(B
				m = position(r, c);
				n = bc(r, c).Re_position_of_variable(-l, UPPER, DEPTH);
				M_all[i][2 * m + n] = - 1.;
				F_all[i] = F_all[i] + bc(r, c).Re_variable(-l, UPPER,
														   DEPTH);
				i++;
			}
		}
	}

	if (i != n_cn * 2) {
		Eprint(__FILE__, __LINE__);	Oprint(); exit(1);
	}

//	printf("M_all\n");
//	for (i = 0; i < 2 * n_cn; i++) {
//		for (j = 0; j < 2 * n_cn; j++) {
//			printf("%f ", M_all[i][j]);
//		}
//		printf("\n%f\n", F_all[i]);
//	}
//	printf("\n");

}

void FldDynaNR::Solve_equation(void)
{
	// $B%,%&%9>C5nK!(B ($B%T%\%C%HA*Br$"$j(B)

	int    	i, j, k, imax;
	double 	p, q, r, s;

	// $B%P%i%s%7%s%0(B
	
	for (i = 0; i < n_cn * 2; i++) {
		p = 0. ;
		for (j = 0; j < n_cn * 2; j++) {
	   		if (fabs(M_all[i][j]) > p) {
				p = fabs(M_all[i][j]);
			}
		}							
		for (j = 0; j < n_cn * 2; j++) {
			if (p != 0) {
	   			M_all[i][j] = M_all[i][j] / p;
			}
		}	
		F_all[i] = F_all[i] / p;
	}		

	// $B%T%\%C%HA*Br(B ($BItJ,A*Br(B)

	for (k = 0; k < n_cn * 2 - 1; k++) {
		p = 0. ;
		imax = k;
		for (i = k; i < n_cn * 2; i++) {
			if (fabs(M_all[i][k]) > p) {
				p = fabs(M_all[i][k]);
				imax = i;
			}
		}
		if (imax != k) {
			for (i = k; i < n_cn * 2; i++) {
				q = M_all[k][i];
				M_all[k][i] = M_all[imax][i];
				M_all[imax][i] = q;
			}
			r = F_all[k];
			F_all[k] = F_all[imax];
			F_all[imax] = r;
		}		

	// $BA]$-$@$77W;;(B
		
	  	for (j = k + 1; j < n_cn * 2; ++j) {
			if (M_all[k][k] == 0) {
				Eprint(__FILE__, __LINE__);
				fprintf(stderr, "current_time = %ld\n", current_time);
				Oprint(); exit(1);			
			}
			M_all[k][j] = M_all[k][j] / M_all[k][k];
		}
		F_all[k] = F_all[k] / M_all[k][k];

		for (i = k + 1; i < n_cn * 2; ++i) {
			for (j = k + 1; j < n_cn * 2; ++j)
				M_all[i][j] = M_all[i][j] - M_all[i][k] * M_all[k][j];
				F_all[i] = F_all[i] - M_all[i][k] * F_all[k];
			
		}
	}
	
	// $B8e?JBeF~(B

 	x_all[n_cn * 2 - 1] = F_all[n_cn * 2 - 1] 
							/ M_all[n_cn * 2 - 1][n_cn * 2 - 1];
	for (k = n_cn * 2 - 2; k >= 0; --k) {
		s = F_all[k];
		for (j = k + 1; j < 2 * n_cn; ++j) {
			s = s - M_all[k][j] * x_all[j];
		}
		x_all[k] = s;
	}
}

void FldDynaNR::Re_a_and_b(size_t r, size_t c, int s, Vdary& v)
{
	int i;
	Boolean flag = NO;
	for (i = 0; i < n_lower2; i++) {
		if (r == qhr[i].row && c == qhr[i].col && s == qhr[i].segnum) {
			v[0] = qhr[i].a;
			v[1] = qhr[i].b;
			v[2] = qhr[i].zlow;
			flag = YES; break;
		}
	}
	if (flag == NO) {
		Eprint(__FILE__, __LINE__);
		// Cannot return a and b in Q-H relation.
		// FldDynaNR::Re_a_and_b
		Oprint(); exit(1);
	}
}

Boolean FldDynaNR::Renewal(void)
{
	int i, j, k;
	size_t r, c;
	Vdary v_cn;
	Boolean flag = YES;

	for (r = 0; r < n_row; r++) {
		for (c = 0; c < n_col; c++) {
			i = bc(r, c).Re_comp_node();
			v_cn.Resize(i * 2);
			k = position(r, c);
			for (j = 0; j < i; j++) {
				v_cn[j * 2] = x_all[k * 2];
				v_cn[j * 2 + 1] = x_all[k * 2 + 1];
				k++;
			}
			bc(r, c).Renewal(v_cn);
		}
	}

//	for (i = 0; i < n_cn; i++) {
//		printf("Qn = %f, Hn = %f\n", x_all[2 * i], x_all[2 * i + 1]);
//	}

	for (i = 0; i < n_cn; i++) {
		if (fabs(x_all[i * 2]) > Qlevel * EPSQ || fabs(x_all[i * 2 + 1])
			> Hlevel * EPSH) {
			flag = NO;
		}
	}
 	return flag;
}

Boolean FldDynaNR::Calculate(void)
{
	if (sflag == VARIED) {
		Calc_discharge();
		Calc_depth();
		time_step = 1000000;
		printf("Calc is OK.\n");
		return YES;
	}

	time_step = ((predt * 2 < t_step) ? predt * 2 : t_step);

	int i, j, k, l;
	size_t r, c;
	double lat;
	Vdary qup, hlow;

	// $BN.NL5Z$S?e?<$N=i4|$N?dDjCM$r@_Dj$9$k(B

	Boolean converge;
	int counter;
	int iteration;

	// time_step $B$rH>J,$K$9$k$H$$$&<jB3$-$r(B 1/16 $B$K$J$k$^$G7+$jJV$9(B
	for (iteration = 0; iteration < 5; iteration++) {

		k = l = 0;

		// $BN.F~NLEy$N<hF@(B
		for (r = 0; r < n_row; r++) {
			for (c = 0; c < n_col; c++) {	// $BB&J}N.F~NL$r<u$1<h$k(B
				if (((FldSlope*)object_array[0])->Re_lat(r, c,
						current_time + time_step, lat) != YES) {
					Eprint(__FILE__, __LINE__);
					// Cannot get lateral inflow at next time.
					// FldDynaNR::Calculate
					Oprint(); exit(1);
				}

				i = bc(r, c).Re_ur();
				qup.Resize(i);
				for (j = 0; j < i; j++) {
					qup[j] = rp_up[k].Get_data(current_time +
											   time_step);
					k++;
				}

				i = bc(r, c).Re_lr1();
				hlow.Resize(i);
				for (j = 0; j < i; j++) {
					hlow[j] = rp_low[l].Get_data(current_time +
											   time_step);
					l++;
				}
				bc(r, c).Ready_for_calculate(lat, qup, hlow);
			}
		}

		counter = 0;

		do {
			Make_equation();
			Solve_equation();
			counter++;
		} while ((converge = Renewal()) == NO && counter < 20);

		if (converge == YES) break;
		else {
			if (time_step == 1) {		// time_step = 1 $B$G$bBLL\$@$C$?(B
				Eprint(__FILE__, __LINE__);
				// cannot converge calculation.
				Oprint(); exit(1);
			}
			double dt = (double)time_step * 0.5;
			time_step = (int)dt;
			if (iteration != 4) {
				printf("Retry using time_step = %ld\n", time_step);
			}
		}
	}

	// time_step $B$r%9%?!<%H;~$N(B 16 $BJ,$N(B 1 $B$K$7$?$,<}B+$7$J$+$C$?(B
	if (iteration == 5) {
		Eprint(__FILE__, __LINE__);
		// cannot converge calculation.
		Oprint(); exit(1);
	}

	// $B7W;;=*N;8e$N=hM}(B

	for (r = 0; r < n_row; r++) {		// $B>uBVNL$N99?7(B
		for (c = 0; c < n_col; c++) {
			bc(r, c).Reset_variable();
		}
	}

	int num;
	time_t t;
	Vdary Qlr;
	Vdary* Qall;
	Vdary* Hall;

//======================================================================
//	$B7W;;7k2L$O:G2<N.C<N.NL$O?^3THV9f$N<c$$=g$G2OF;6hJ,HV9f$N<c$$=g$K(B
//	$B3F%;%k$G$N>uBVNL$OEPO?=g(B($B2OF;LVA4BN$K4X$9$k%U%!%$%k$K$*$1$k?^3THV9f(B
//	$B5Z$S!"3F%;%kFbIt$K4X$9$k%U%!%$%k$K$*$1$k2OF;6hJ,HV9f$N=g(B)$B$K$J$j$^$9!#(B
//======================================================================

	if (current_time % 60 < time_step) {
		j = 0;
		for (r = 0; r < n_row; r++) {
			for (c = 0; c < n_col; c++) {
				num = bc(r, c).Re_lr1() + bc(r, c).Re_lr2();
				Qlr	.Resize(num);
				bc(r, c).Re_lowest_q(Qlr);
				for (i = 0; i < num; i++) {
					sp[j].Send_data(current_time + time_step, Qlr[i]);
					j++;
				}
			}
		}

		if (j != (n_lower1 + n_lower2)) {
			Eprint(__FILE__, __LINE__);
			Oprint(); exit(1);
		}
		l = 0;
		for (i = 0; i < n_parts; i++) {
			r = cn[i].row;
			c = cn[i].col;
			j = bc(r, c).Re_num_of_seg();
			Qall = new Vdary[j];
			Hall = new Vdary[j];
			if (bc(r, c).Output_variable(Qall, Hall, NEXT) == YES) {
				for (k = 0; k < j; k++) {
					sp_q[l].Send_data(current_time + time_step,
										  Qall[k]);
					sp_h[l].Send_data(current_time + time_step,
										  Hall[k]);
					l++;
				}
			}
			else {
				Eprint(__FILE__, __LINE__);
				Oprint(); exit(1);
			}
			delete [] Qall;
			delete [] Hall;
		}
	}

//	printf("current_time = %8ld ", current_time);
	if ((current_time % 3600) < time_step) {
		t = current_time / 3600;
		if (t == 0 || t == 1) {
			printf("%4ld hour\n", t);
		}
		else {
			printf("%4ld hours\n", t);
		}
	}
//	printf("\n");
	predt = time_step;
	return YES;
} 

void FldDynaNR::Calc_discharge(void)
{
	int i, j, s;
	size_t r, c, rr, cc;
	double d;
	Vdary q;
	Boolean flag, all_flag;
	Boolean* calc_flag;

	int counter = 0;
	do {
		all_flag = YES;
		for (r = 0; r < n_row; r++) {
			for (c = 0; c < n_col; c++) {
				if (bc(r, c).Re_Qflag() == NO) {
					i = sn_up(r, c).Size();
					calc_flag = new Boolean[i];
					q.Resize(i);
					for (j = 0; j < i; j++) {
						rr = bp_in(r, c).row[j];
						cc = bp_in(r, c).col[j];
						s = bp_in(r, c).segnum[j];
						if (bc(rr, cc).Re_q(s, d) == YES) {
							calc_flag[j] = YES;
							q[j] = d;
						}
						else {
							calc_flag[j] = NO;
						}
					}
					flag = YES;
					for (j = 0; j < i; j++) {
						if (calc_flag[j] == NO) {
							flag = NO; break;
						}
					}
					if (flag == YES) {
						if (bc(r, c).Calc_discharge(sn_up(r, c), q) !=
							YES) {
							Eprint(__FILE__, __LINE__);
							// Cannot calculate discharge.
							// FldDynaNR::Calc_discharge
							Oprint(); exit(1);
						}
					}
					else all_flag = NO;
					delete [] calc_flag;
				}
			}
		}
		counter++;
		if (counter > 100000) {
			fprintf(stderr, "Cannot calculate discharge.\n");
			exit(1);
		}
	} while (all_flag == NO);
}

void FldDynaNR::Calc_depth(void)
{
	int i, j, s;
	size_t r, c, rr, cc;
	double d;
	Vdary h;
	Boolean flag, all_flag;
	Boolean* calc_flag;

	do {
		all_flag = YES;
		for(r = 0; r < n_row; r++) {
			for (c = 0; c < n_col; c++) {
				if (bc(r, c).Re_Hflag() == NO) {
					i = sn_low(r, c).Size();
					calc_flag = new Boolean[i];
					h.Resize(i);
					for (j = 0; j < i; j++) {
						rr = bp_out(r, c).row[j];
						cc = bp_out(r, c).col[j];
						s = bp_out(r, c).segnum[j];
						if (bc(rr, cc).Re_h(s, d) == YES) {
							calc_flag[j] = YES;
							h[j] = d;
						}
						else {
							calc_flag[j] = NO;
						}
					}
					flag = YES;
					for (j = 0; j < i; j++) {
						if (calc_flag[j] == NO) {
							flag = NO; break;
						}
					}
					if (flag == YES) {
						if (bc(r, c).Calc_depth(sn_low(r, c), h) !=
							YES) {
							Eprint(__FILE__, __LINE__);
							// Cannot calculate depth.
							// FldDynaNR::Calc_depth
							Oprint(); exit(1);
						}
					}
					else all_flag = NO;
					delete [] calc_flag;
				}
			}
		}
	} while (all_flag == NO);
}

	

FldDynaNR::~FldDynaNR(void)
{
	int i;

	for (i = 0; i < n_cn * 2; i++) {
		delete [] M_all[i];
		delete [] M_fixed[i];
	}
	for (i = 0; i < n_max; i++) {
		delete [] M_cell[i];
	}
	delete [] M_all;
	delete [] M_fixed;
	delete [] M_cell;
	delete [] bp;
	delete [] cn;
	delete [] qhr;
	delete [] rp_up;
	delete [] rp_low;
	delete [] sp;
	delete [] sp_q;
	delete [] sp_h;
}
