// segmentNR.cc

//----------------------------------------------------------------------
//	SegmentNR		$BC10l2OF;%/%i%9(B
//----------------------------------------------------------------------

#include "segmentNR.h"	// $B%/%i%9(B SegmentNR $B$NDj5A(B

Boolean SegmentNR::Init(int snum, int unode, int lnode)
{
	if (snum <= 0 || unode < 0 || lnode < 0) return NO;

	segnum = snum;
	upnode = unode;
	lownode = lnode;

	return YES;
}

SegmentNR::SegmentNR(int snum, int unode, int lnode)
{
	Init(snum, unode, lnode);
}

Boolean SegmentNR::Set_parameter(FILE* fp, time_t time_step)
{
	int i;
	char str[1024];
	char* pstr = str;

	if (fgets(str, 1024, fp) == NULL) return NO;
	else if (sscanf(str, "%lf %lf %lf %lf %lf %d", &L, &zup, &zlow, &man,
			   &theta, &n_cs) != 6) {
		return NO;
	}

	io = (zup - zlow) / L;

	Qn.Resize(n_cs);
	Hn.Resize(n_cs);
	Qc.Resize(n_cs);
	Hc.Resize(n_cs);

	dq.Resize(n_cs);
	dh.Resize(n_cs);
	dx.Resize(n_cs - 1);
	F.Resize(2 * n_cs - 2);

	An.Resize(n_cs);
	A0n.Resize(n_cs);
	Ifn.Resize(n_cs);
	Ac.Resize(n_cs);
	A0c.Resize(n_cs);
	Ifc.Resize(n_cs);

	b.Resize(n_cs);
	b0.Resize(n_cs);

	// 	$B7W;;CGLL$N0LCV$+$i6u4V4V3V$r7hDj$9$k(B

	double dx_first, dx_second;

	if (fgets(str, 1024, fp) == NULL) return NO;
	else {
		while (*pstr == ' ' || *pstr == '\t') pstr++;
		if (sscanf(pstr, "%lf", &dx_first) != 1) return NO;
		while (*pstr != ' ' && *pstr != '\t') pstr++;
		for (i = 0; i < n_cs - 1; i++) {
			while (*pstr == ' ' || *pstr == '\t') pstr++;
			if (sscanf(pstr, "%lf", &dx_second) != 1) return NO;
			while (*pstr != ' ' && *pstr != '\t') pstr++;
			if ((dx_second - dx_first) < 0.) return NO;
			dx[i] = dx_second - dx_first;
			dx_first = dx_second;
		}
	}

	// $B2OF;I}(B(active and inactive)

	double d;

	if (fgets(str, 1024, fp) == NULL) return NO;
	else {
		pstr = str;
		for (i = 0; i < n_cs; i++) {
			while (*pstr == ' ' || *pstr == '\t') pstr++;
			if (sscanf(pstr, "%lf", &d) != 1) return NO;
			if (d < 0.) return NO;
			b[i] = d;
			while (*pstr != ' ' && *pstr != '\t') pstr++;
		}
	}

	if (fgets(str, 1024, fp) == NULL) return NO;
	else {
		pstr = str;
		for (i = 0; i < n_cs; i++) {
			while (*pstr == ' ' || *pstr == '\t') pstr++;
			if (sscanf(pstr, "%lf", &d) != 1) return NO;
			if (d < 0.) return NO;
			b0[i] = d;
			while (*pstr != ' ' && *pstr != '\t') pstr++;
		}
	}

	// $B%d%3%S9TNs(B

	J = new double*[2 * n_cs - 2];
	for (i = 0; i < 2 * n_cs - 2; i++) {
		J[i] = new double[4];
	}

	J_fixed = new double*[n_cs - 1];
	for (i = 0; i < n_cs - 1; i++) {
		J_fixed[i] = new double[2];
	}

	start_t_step = time_step;

	for (i = 0; i < n_cs - 1; i++) {
		J_fixed[i][0] = 0.5 * (b[i] + b0[i]) * (dx[i] /
													(double)time_step);
		J_fixed[i][1] = 0.5 * (b[i + 1] + b0[i + 1]) * (dx[i] /
													  (double)time_step);
	}

	return YES;
}

void SegmentNR::Set_init_for_varied(double qup, double qlat)
{
	// $BITEyN.$N7W;;$N$?$a$N=i4|>uBVNL$r@_Dj$7$^$9!#(B($B>eN.C<MQ(B)
	// qup : $B>eN.C<$G$NN.NL(B
	// qlat : $BC10LI}$"$?$j$NB&J}N.F~NL(B

	int i;

	Qc[0] = qup;
	Qlatc = qlat;
	
	for (i = 0; i < n_cs - 1; i++) {
		Qc[i + 1] = Qc[i] + Qlatc * dx[i];
	}
	Qflag = YES;		// $BN.NL$O7W;;:Q$_(B
	Hflag = NO;			// $B?e?<$OL$7W;;(B
}

void SegmentNR::Set_init_for_varied(double qlat)
{
	// $BITEyN.$N7W;;$N$?$a$N=i4|>uBVNL$r@_Dj$7$^$9!#(B($B>eN.C<0J30MQ(B)
	// qlat : $BC10LI}$"$?$j$NB&J}N.F~NL(B

	Qlatc = qlat;
	Qflag = NO;			// $BN.NL$OL$7W;;(B
	Hflag = NO;			// $B?e?<$OL$7W;;(B
}

Boolean SegmentNR::Set_init_for_runoff(FILE* fp, double lat)
{
	int i;
	double d;
	char str[1024];
	char* pstr = str;

	Qlatc = lat;

	if (fgets(str, 1024, fp) == NULL) return NO;
	else {
		for (i = 0; i < n_cs; i++) {
			while (*pstr == ' ' || *pstr == '\t') pstr++;
			if (sscanf(pstr, "%lf", &d) != 1) return NO;
			Qc[i] = d;
			while (*pstr != ' ' && *pstr != '\t') pstr++;
		}	
	}

	if (fgets(str, 1024, fp) == NULL) return NO;
	else {
		pstr = str;
		for (i = 0; i < n_cs; i++) {
			while (*pstr == ' ' || *pstr == '\t') pstr++;
			if (sscanf(pstr, "%lf", &d) != 1) return NO;
			if (d <= 0.) return NO;
			Hc[i] = d;
			while (*pstr != ' ' && *pstr != '\t') pstr++;
		}	
	}

	// $BCGLL@Q(B(active, inactive)$B!&K`;$8{G[$N7W;;(B

	for (i = 0; i < n_cs; i++) {
		Ac[i] = b[i] * Hc[i];
		A0c[i] = b0[i] * Hc[i];

		if (Hc[i] == 0.) {
			printf("Cannot compute Ifc[%d], Hc[%d] == 0.\n", i, i);
			printf("SegmentNumber = %d\n", segnum);
			exit(1);
		}
		Ifc[i] = man * man * fabs(Qc[i]) * Qc[i] / (Ac[i] * Ac[i] *
													pow(Hc[i], 4./3.));
	}
	return YES;
}

Boolean SegmentNR::Save_terminal_state(FILE* fp)
{
	int i;

	if (fp == NULL) return NO;
	else {
		for (i = 0; i < n_cs; i++) {
			fprintf(fp, "%f ", Qc[i]);
		}
	}

	fprintf(fp, "\n");

	for (i = 0; i < n_cs; i++) {
		fprintf(fp, "%f ", Hc[i]);
	}

	fprintf(fp, "\n");
	return YES;
}

double SegmentNR::Re_node_variable(ULFlag ul, QHFlag qh, CNFlag s)
{
	// $B;XDj$5$l$?>l=j$G$NN.NL$^$?$O?e?<$rJV$9(B

	if (s == CURRENT) {
		if (ul == UPPER) {
			if (qh == DISCHARGE) return Qc[0];
			else if (qh == DEPTH) return Hc[0];
		}
		else if (ul == LOWER) {
			if (qh == DISCHARGE) return Qc[n_cs - 1];
			else if (qh == DEPTH) return Hc[n_cs - 1];
		}
	}
	else if (s == NEXT) {
		if (ul == UPPER) {
			if (qh == DISCHARGE) return Qn[0];
			else if (qh == DEPTH) return Hn[0];
		}
		else if (ul == LOWER) {
			if (qh == DISCHARGE) return Qn[n_cs - 1];
			else if (qh == DEPTH) return Hn[n_cs - 1];
		}
	}

	printf("Invalid parameter.\n");
	printf("SegmentNR::Re_node_variable\n");
	exit(1);
}

void SegmentNR::First_estimation(double lat)
{
	// $BN.NL5Z$S?e?<$N=i4|$N?dDjCM$r@_Dj$9$k(B

	Qn = Qc;
	Hn = Hc;
	An = Ac;
	A0n = A0c;
	Ifn = Ifc;
	Qlatn = lat;
}

void SegmentNR::Set_lower_condition(double hlow)
{
	// $B2<N.C<6-3&>r7o$r@_Dj$9$k(B
	// hlow : $B2<N.C<6-3&$G$N?e0L(B

	Hn[n_cs - 1] = hlow - zlow;		// $B?e?<$KD>$9(B
	An[n_cs - 1] = b[n_cs - 1] * Hn[n_cs - 1];
	A0n[n_cs - 1] = b0[n_cs - 1] * Hn[n_cs - 1];

	if (Hn[n_cs - 1] == 0.) {
		printf("Cannot compute Ifn[%d], Hn[%d] == 0.\n", n_cs - 1, n_cs
			   - 1);
		printf("SegmentNumber = %d\n", segnum);
		printf("Seg_lower_condition\n");
		exit(1);
	}
	Ifn[n_cs - 1] = man * man * fabs(Qn[n_cs - 1]) * Qn[n_cs - 1] /
		(An[n_cs - 1] * An[n_cs - 1] * pow(Hn[n_cs - 1], 4./3.));
}

Boolean SegmentNR::Upnode(int unode)
{
	if (unode == upnode) return YES;
	else return NO;
}

Boolean SegmentNR::Lownode(int lnode)
{
	if (lnode == lownode) return YES;
	else return NO;
}

void SegmentNR::Make_equation(time_t time_step)
{
	int i;
	double d1, d2, d3, d4, d5;

	for (i = 0; i < n_cs - 1; i++) {
		J[2 * i][0] = - theta;
		J[2 * i][2] = theta;
		if (time_step == start_t_step) {
			J[2 * i][1] = J_fixed[i][0];
			J[2 * i][3] = J_fixed[i][1];
		}
		else {
			J[2 * i][1] = 0.5 * (b[i] + b0[i]) * (dx[i] /
												  (double)time_step);
			J[2 * i][3] = 0.5 * (b[i+1] + b0[i+1]) * (dx[i] /
													  (double)time_step);
		}

		d1 = 0.5 * (dx[i] / (double)time_step);
		d2 = 2. * theta * (Qn[i] / An[i]);
		d3 = 0.5 * G * (theta * (An[i] + An[i+1]) + (1. - theta) *
						(Ac[i] + Ac[i+1]));
//		d4 = dx[i] * theta * (Ifn[i] / Qn[i]);
		d4 = dx[i] * theta * ((man * man * fabs(Qn[i])) / (b[i] * b[i] * 
												   pow(Hn[i], 10./3.)));

		J[2 * i + 1][0] = d1 - d2 + d3 * d4;

		d1 = theta * ((Qn[i] * Qn[i]) / (An[i] * Hn[i]));
		d2 = theta * (Hn[i+1] - Hn[i]) + (1. - theta) *	(Hc[i+1] -
														 Hc[i]);
		d3 = 0.5 * dx[i] * (theta * (Ifn[i+1] + Ifn[i]) 
							+ (1. - theta) * (Ifc[i+1] + Ifc[i]));
		d4 = theta * (An[i] + An[i+1]) + (1. - theta) * (Ac[i] +
														 Ac[i+1]);
//		d5 = 1. + (5./3.) * dx[i] * Ifn[i] / Hn[i];
		d5 = 1. + (5./3.) * dx[i] * ((man * man * fabs(Qn[i]) * Qn[i]) /
									 (b[i] * b[i] * pow(Hn[i], 13./3.)));

		J[2 * i + 1][1] = d1 + 0.5 * G * theta * (b[i] * (d2 + d3 -
											  dx[i] * io) - d4 * d5);

		d1 = 0.5 * (dx[i] / (double)time_step);
		d2 = 2. * theta * (Qn[i+1] / An[i+1]);
		d3 = 0.5 * G * (theta * (An[i] + An[i+1]) + (1. - theta) *
						(Ac[i] + Ac[i+1]));
//		d4 = dx[i] * theta * (Ifn[i+1] / Qn[i+1]);
		d4 = dx[i] * theta * ((man * man * fabs(Qn[i+1]))
						  / (b[i+1] * b[i+1] * pow(Hn[i+1], 10./3.)));

		J[2 * i + 1][2] = d1 + d2 + d3 * d4;

		d1 = theta * ((Qn[i+1] * Qn[i+1]) / (An[i+1] * Hn[i+1]));
		d2 = theta * (Hn[i+1] - Hn[i]) + (1. - theta) *	(Hc[i+1] -
														 Hc[i]);
		d3 = 0.5 * dx[i] * (theta * (Ifn[i+1] + Ifn[i]) 
							+ (1. - theta) * (Ifc[i+1] + Ifc[i]));
		d4 = theta * (An[i] + An[i+1]) + (1. - theta) * (Ac[i] +
														 Ac[i+1]);
//		d5 = 1. - (5./3.) * dx[i] * Ifn[i+1] / Hn[i+1];
		d5 = 1. - (5./3.) * dx[i] * ((man * man * fabs(Qn[i+1]) *
									  Qn[i+1]) /
							 (b[i+1] * b[i+1] * pow(Hn[i+1], 13./3.)));

		J[2 * i + 1][3] = - d1 + 0.5 * G * theta * (b[i+1] * 
									(d2 + d3 - dx[i] * io) + d4 * d5);

		d1 = theta * (Qn[i+1] - Qn[i] - Qlatn * dx[i])
			+ (1 - theta) * (Qc[i+1] - Qc[i] - Qlatc * dx[i]);
		d2 = 0.5 * (dx[i] / (double)time_step) 
				* (((An[i+1] + A0n[i+1]) + (An[i] + A0n[i])) 
				   - ((Ac[i+1] + A0c[i+1]) +(Ac[i] + A0c[i])));

		F[2 * i] = - d1 - d2;

		d1 = 0.5 * (dx[i] / (double)time_step) * ((Qn[i+1] + Qn[i]) -
												  (Qc[i+1] + Qc[i]));
		d2 = theta * (((Qn[i+1] * Qn[i+1]) / An[i+1])
				- ((Qn[i] * Qn[i]) / An[i]))
			+ (1 - theta) * (((Qc[i+1] * Qc[i+1]) / Ac[i+1])
							 - ((Qc[i] * Qc[i]) / Ac[i]));
		d3 = 0.5 * G * (theta * (An[i+1] + An[i])
						+ (1 - theta) * (Ac[i+1] + Ac[i]));
		d4 = theta * (Hn[i+1] - Hn[i]) + (1 - theta) * (Hc[i+1] -
														Hc[i]);
		d5 = 0.5 * dx[i] * (theta * (Ifn[i+1] + Ifn[i])
							+ (1 - theta) * (Ifc[i+1] + Ifc[i]));

		F[2 * i + 1] = - d1 - d2 - d3 * (d4 + d5 - dx[i] * io);
	}

//	printf("MakeSeg %d\n", segnum);
//	for (i = 0; i < 2 * n_cs - 2; i++) {
//		printf("%f %f %f %f || %15.10f\n", J[i][0], J[i][1],
//			   J[i][2], J[i][3], F[i]);
//	}
//	printf("\n");

}

Boolean SegmentNR::Sweep_equation(double**& c)
{
	int i;

	for (i = 0; i < n_cs - 1; i++) {
		if (J[2 * i][2] == 0. || J[2 * i + 1][3] == 0.) return NO;
	}

	// $BA]$-$@$77W;;(B

	double a[2];

	for (i = 0; i < n_cs - 2; i++) {

	// Part 1

	J[2 * i][0] = J[2 * i][0] / J[2 * i][2];
	J[2 * i][1] = J[2 * i][1] / J[2 * i][2];
	J[2 * i][3] = J[2 * i][3] / J[2 * i][2];
	F[2 * i] = F[2 * i] / J[2 * i][2];
	J[2 * i][2] = 1.;

	J[2 * i + 1][0] = J[2 * i + 1][0] - J[2 * i][0] *
	  J[2 * i + 1][2];
	J[2 * i + 1][1] = J[2 * i + 1][1]
	  - J[2 * i][1] * J[2 * i + 1][2];
	J[2 * i + 1][3] = J[2 * i + 1][3]
	  - J[2 * i][3] * J[2 * i + 1][2];
	F[2 * i + 1] = F[2 * i + 1] - F[2 * i] * J[2 * i + 1][2];
	J[2 * i + 1][2] = 0.;

	a[0] = J[2 * i + 2][1] - J[2 * i][3] * J[2 * i + 2][0];
	F[2 * i + 2] = F[2 * i + 2] - F[2 * i] * J[2 * i + 2][0];
	J[2 * i + 2][1] = - (J[2 * i][1] * J[2 * i + 2][0]);
	J[2 * i + 2][0] = - (J[2 * i][0] * J[2 * i + 2][0]);

	a[1] = J[2 * i + 3][1] - J[2 * i][3] * J[2 * i + 3][0];
	F[2 * i + 3] = F[2 * i + 3] - F[2 * i] * J[2 * i + 3][0];
	J[2 * i + 3][1] = - (J[2 * i][1] * J[2 * i + 3][0]);
	J[2 * i + 3][0] = - (J[2 * i][0] * J[2 * i + 3][0]);

	// Part 2

	J[2 * i + 1][0] = J[2 * i + 1][0] / J[2 * i + 1][3];
	J[2 * i + 1][1] = J[2 * i + 1][1] / J[2 * i + 1][3];
	F[2 * i + 1] = F[2 * i + 1] / J[2 * i + 1][3];
	J[2 * i + 1][3] = 1.;

	J[2 * i][0] = J[2 * i][0] - J[2 * i + 1][0] * J[2 * i][3];
	J[2 * i][1] = J[2 * i][1] - J[2 * i + 1][1] * J[2 * i][3];
	F[2 * i] = F[2 * i] - F[2 * i + 1] * J[2 * i][3];
	J[2 * i][3] = 0.;

	J[2 * i + 2][0] = J[2 * i + 2][0] - J[2 * i + 1][0] * a[0];
	J[2 * i + 2][1] = J[2 * i + 2][1] - J[2 * i + 1][1] * a[0];
	F[2 * i + 2] = F[2 * i + 2] - F[2 * i + 1] * a[0];

	J[2 * i + 3][0] = J[2 * i + 3][0] - J[2 * i + 1][0] * a[1];
	J[2 * i + 3][1] = J[2 * i + 3][1] - J[2 * i + 1][1] * a[1];
	F[2 * i + 3] = F[2 * i + 3] - F[2 * i + 1] * a[1];
  }

	// $B:G2<N.$N#29`$NA]$-$@$7(B

	J[2 * n_cs - 4][0] = J[2 * n_cs - 4][0] / J[2 * n_cs - 4][2];
	J[2 * n_cs - 4][1] = J[2 * n_cs - 4][1] / J[2 * n_cs - 4][2];
	J[2 * n_cs - 4][3] = J[2 * n_cs - 4][3] / J[2 * n_cs - 4][2];
	F[2 * n_cs - 4] = F[2 * n_cs - 4] / J[2 * n_cs - 4][2];
	J[2 * n_cs - 4][2] = 1.;

	J[2 * n_cs - 3][0] = J[2 * n_cs - 3][0] - J[2 * n_cs - 4][0]
	  * J[2 * n_cs - 3][2];
	J[2 * n_cs - 3][1] = J[2 * n_cs - 3][1] - J[2 * n_cs - 4][1]
	  * J[2 * n_cs - 3][2];
	J[2 * n_cs - 3][3] = J[2 * n_cs - 3][3] 
	  - J[2 * n_cs - 4][3] * J[2 * n_cs - 3][2];
	F[2 * n_cs - 3] = F[2 * n_cs - 3] - F[2 * n_cs - 4]
	  * J[2 * n_cs - 3][2];
	J[2 * n_cs - 3][2] = 0.;

	J[2 * n_cs - 3][0] = J[2 * n_cs - 3][0] / J[2 * n_cs - 3][3];
	J[2 * n_cs - 3][1] = J[2 * n_cs - 3][1] / J[2 * n_cs - 3][3];
	F[2 * n_cs - 3] = F[2 * n_cs - 3] / J[2 * n_cs - 3][3];
	J[2 * n_cs - 3][3] = 1.;

	J[2 * n_cs - 4][0] = J[2 * n_cs - 4][0] - J[2 * n_cs - 3][0]
	  * J[2 * n_cs - 4][3];
	J[2 * n_cs - 4][1] = J[2 * n_cs - 4][1] - J[2 * n_cs - 3][1]
	  * J[2 * n_cs - 4][3];
	F[2 * n_cs - 4] = F[2 * n_cs - 4] - F[2 * n_cs - 3]
	  * J[2 * n_cs - 4][3];
	J[2 * n_cs - 4][3] = 0.;

	c[0][0] = J[2 * n_cs - 4][0];
	c[0][1] = J[2 * n_cs - 4][1];
	c[0][2] = F[2 * n_cs - 4];
	c[1][0] = J[2 * n_cs - 3][0];
	c[1][1] = J[2 * n_cs - 3][1];
	c[1][2] = F[2 * n_cs - 3];

	return YES;
}

void SegmentNR::Renewal(double* data)
{
	int i;
	double d, lambda;

	dq[0] = data[0];
	dh[0] = data[1];

	for (i = 1; i < n_cs; i++) {
		dq[i] = F[2 * i - 2] - J[2 * i - 2][0] * dq[0] 
			- J[2 * i - 2][1] * dh[0];

		dh[i] = F[2 * i - 1] - J[2 * i - 1][0] * dq[0]
			- J[2 * i - 1][1] * dh[0];
	}

	lambda = 1.0;

	for (i = 0; i < n_cs; i++) {
		if (dh[i] < 0.) {
		d = - fabs(Hn[i] - EPS) / dh[i];
			if (lambda > d) {
				lambda = d;
			}
		}
	}

//	if (lambda < 0.1) lambda = 0.1;

	for (i = 0; i < n_cs; i++) {
// 		Qn[i] = Qn[i] + dq[i];
//		Hn[i] = fabs(Hn[i] + 0.7 * dh[i]);

		// modified by Y. Ichikawa
//		Qn[i] = Qn[i] + 0.1 * dq[i];
//		Hn[i] = fabs(Hn[i] + 0.1 * dh[i]);

		d = Hn[i];

 		Qn[i] = Qn[i] + dq[i];
//		Hn[i] = fabs(Hn[i] + lambda * dh[i]);
		Hn[i] = Hn[i] + lambda * dh[i];

		An[i] = b[i] * Hn[i];
		A0n[i] = b0[i] * Hn[i];

		if (Hn[i] <= 0.) {
			printf("Cannot compute Ifn[%d]. preHn = %f, Hn[%d] = %f\n",
				   i, d, i, Hn[i]);
			printf("lambda = %f, dh = %f\n", lambda, dh[i]);
			printf("SegmentNumber = %d\n", segnum);
			printf("Renewal\n");
			exit(1);
		}

		Ifn[i] = man * man * fabs(Qn[i]) * Qn[i] / (An[i] * An[i] *
			pow(Hn[i], 4./3.));
	}
}

void SegmentNR::Reset_variable(void)
{
	Qc = Qn;
	Hc = Hn;
	Ac = An;
	A0c = A0n;
	Ifc = Ifn;
	Qlatc = Qlatn;
}

Boolean SegmentNR::Output_variable(Vdary& Qall, Vdary& Hall,
										 CNFlag s)
{
	int i;

	if (Qall.Size() != n_cs || Hall.Size() != n_cs) return NO;

	if (s == CURRENT) {
		for (i = 0; i < n_cs; i++) {
			Qall[i] = Qc[i];
			Hall[i] = Hc[i];
		}
	}
	else if (s == NEXT) {
		for (i = 0; i < n_cs; i++) {
			Qall[i] = Qn[i];
			Hall[i] = Hn[i];
		}
	}
	else return NO;
	return YES;
}

void SegmentNR::Calc_discharge(double qup)
{
	int i;

	if (Qflag == YES) return;	// $B7W;;:Q$_(B

	Qc[0] = qup;
	for (i = 0; i < n_cs - 1; i++) {
		Qc[i + 1] = Qc[i] + Qlatc * dx[i];
	}
	Qflag = YES;
}

Boolean SegmentNR::Calc_depth(double hlow, Boolean lowflag)
{
	int i;
	int counter;
	double F, F_h;
	double eps;

	if (Hflag == YES) return YES;	// $B7W;;:Q$_(B

	if (lowflag == YES) {
		Hc[n_cs - 1] = pow(((man * Qc[n_cs - 1]) / (b[n_cs - 1] *
													sqrt(io))), 0.6);
	}
	else {
		Hc[n_cs - 1] = hlow;
	}

	for (i = n_cs - 2; i >= 0; i--) {
		eps = 1.; counter = 0;
		Hc[i] = Hc[i + 1];

		while ((fabs(eps) > 0.000001) && (counter < 1000)) {

			F = pow((Qc[i] / b[i] / Hc[i]), 2.) / 2. / G
				+ Hc[i]
				- man * man * dx[i] / 2. 
					* Qc[i] * Qc[i]
					/ b[i] / b[i] / pow(Hc[i], 10. / 3.)
				- pow((Qc[i + 1] / b[i + 1] / Hc[i + 1]), 2.) / 2. / G
				- Hc[i + 1]
				- man * man * dx[i] / 2. 
					* Qc[i + 1] * Qc[i + 1]	/ b[i + 1] / b[i + 1] / 
					pow(Hc[i + 1], 10. / 3.)
					+ io * dx[i];

			F_h = - Qc[i] * Qc[i] / b[i]
					  / b[i] / G / pow(Hc[i], 3.) + 1.
				  + 5. / 3. * man * man * dx[i] * Qc[i]
					  * Qc[i] / b[i] / b[i] / pow(Hc[i], 13. / 3.);

			eps = - F / F_h;

			if (Hc[i] + eps <= 0.) {
				Hc[i] = fabs(Hc[i] + 0.7 * eps);
			}
			else {
				Hc[i] = Hc[i] + eps;
			}
			counter++;
		}
		if (counter >= 1000) return NO;
	}
	Hflag = YES;
	return YES;
}

Boolean SegmentNR::Re_q(double& q)
{
	if (Qflag == NO) return NO;

	q = Qc[n_cs - 1];
	return YES;
}

Boolean SegmentNR::Re_h(double& h)
{
	if (Hflag == NO) return NO;

	h = Hc[0];
	return YES;
}

SegmentNR::~SegmentNR(void)
{
	int i;

	for (i = 0; i < 2 * n_cs - 2; i++) {
		delete [] J[i];
	}
	for (i = 0; i < n_cs - 1; i++) {
		delete [] J_fixed[i];
	}
	delete [] J;
	delete [] J_fixed;
}
